%DSL Simulate a standard digital subscriber line.
% [Y, H, N] = DSL(X, channelName, P, S, U) returns the received signal
% Y (without channel noise), channel impulse response in H, and
% channel noise in N. The channel noise consists of NEXT noise 
% of U users and AWGN of power P. 
%
% X is the transmitted signal. channelName is a string containing
% the channel name. The channel data should be in a subdirectory 
% "channel". Currently supported channels are the CSA loops from 1 
% to 8. The name format is "csaloop#" where # = {1,2,..,8}. P is the
% AWGN power. S is a subsampling factor. The default length of
% H is 512 samples. A shorter channel can be generated by subsampling
% H by a factor of S. U is the number of users causing NEXT noise.

% Copyright (c) 1999-2000 The University of Texas
% All Rights Reserved.
%  
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%  
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%  
% The GNU Public License is available in the file LICENSE, or you
% can write to the Free Software Foundation, Inc., 59 Temple Place -
% Suite 330, Boston, MA 02111-1307, USA, or you can find it on the
% World Wide Web at http://www.fsf.org.
%  
% Programmers:	Guner Arslan
% Version:        @(#)dsl.m	1.4	07/26/00
% 
% The authors are with the Department of Electrical and Computer
% Engineering, The University of Texas at Austin, Austin, TX.
% They can be reached at arslan@ece.utexas.edu.
% Guner Arslan is also with the Embedded Signal Processing
% Laboratory in the Dept. of ECE., http://anchovy.ece.utexas.edu.

function [y,h,noise] = dsl(x,channelName,noisePower,subsample,users)


% open a figure for progress bar
%[figHndl statusHndl] = setprogbar('Creating Channel Data ...');     
% set path for channel files 
%addpath channels
% load channel data 
channelTime = ['load ',channelName,'.time'];
eval(channelTime)
% put channel data into h and time axis into t
string = ['t = ',channelName,'(:,1); h = ',channelName,'(:,2);'];
eval(string)
% sampling frequency
fsample = 2.208e6/subsample;
% subsample the channel if required
h = h(1:subsample:length(h))/2e6;  
% filter transmit data with channel impulse response
filtX = filter(h,1,x);
% update progress bar
  %updateprogbar(statusHndl,1,4);


% prepare a frequency axis for a filter prototype
f = linspace(0,fsample/2,32);
% desired frequency response of the NEXT coupling filter
Hnext = sqrt((users/49)^0.6 * 1/1.134e13 .* (f.^1.5));
% normalize to unity
f = f./max(f);  
% design IIR NEXT coupling filter with coeffs a and b
[b a] = yulewalk(32,f,Hnext);
% update progress bar
  %updateprogbar(statusHndl,2,4);
% filter a white noise signal with NEXT filter
NEXTinput  = sqrt(cov(x))*randn(size(filtX));
NEXTnoise =  filter( b,a, NEXTinput );
% variance of the NEXT noise
covnext = cov(NEXTnoise);
% update progress bar
  %updateprogbar(statusHndl,3,4);

% generate AWGN with power of AWGNpower
AWGN = sqrt(noisePower).*randn(size(filtX));
% variance of AWGN
covawgn = cov(AWGN);
% add AWGN with NEXT noise
noise = NEXTnoise + AWGN; 
% add the noise to the channel filtered signal
y = filtX + noise;
% calculate obtained SNR
SNRmfb = 10*log10(cov(filtX)./(covawgn+covnext));

% update progress bar
%updateprogbar(statusHndl,4,4);

%close(figHndl);
