/* MIN-ISI ALGORITHM */

#include <stdio.h>
#include "dsplib.h"                      
#include "patch.h"                      
#include "csa1mat.h"							//this file gives us h[]  (channel impulse response)

///////////////////////////////// CONSTANTS /////////////////////////
#define N   		512															//channel length
#define nu  		32															//nu = how many values to select for window
#define Nw  		19															//number of TEQ taps
#define THD			0.000025

///////////////////////////////// Variables /////////////////////////

short i,j,k;
short index;
short window_pwr = 0; 
short max_power = 0;
float oflag = 0;
float errorcode = -1;	

short delta =	0;																	//delta = how many values to wait before choosing window
short first = 0;
short last = 0;
short diff = 0;
 
float h_eff[N];
short h_eff_short[N];  /*used for plotting*/
short h_sqr[N];		                           			// h_sqr[N] 
float h_sqr_float;
short gain[N];
short subchannel[N]; 	/*used for plotting*/

short ipr_sect[N][Nw];														//holds either window or wall values
short Y[Nw][Nw];
short X[Nw][Nw]; 
short Z[Nw][Nw]; 
short temp[N][Nw];																//stores an intermediate matrix multiplication
short subtract[N][4];

short p[Nw];																			//used in inv_of_chol funct
float p_float[Nw];
float sum = 0;
float sum_float = 0;
short q15 = 0;	 
short z_mag = 0;             													/*holds the magnitude of z*/
short lambda = 0; 
short lambda_old = 0;  
short lambda_diff;
short count = 0; 
short start = 0;

short l[Nw];
short z[Nw];
float w_opt[Nw];                    
          
void main (void)
{
 
	for(i=0; i<N; i++)
	  ifft[i] = ifft[i]/1000;

  /* PROFILING STARTS BELOW / START OF TEQ CALCULATIONS */
 
  /* initialize h_sqr  */        
	for(i=0; i<N; i++)
	{   
		h_sqr[i] = h[i] * h[i];
		/* determine last non-zero h value */	
		if((first)&&(h_sqr[i]))
			last = i;                       
		/* determine first non-zero h value */
		if((!first)&&(h_sqr[i]))
			first = i;
	}

	for(index=0; index < N-nu+1; index++){
   	window_pwr = 0; 
   	/* window is of size "nu" */
	  for(k=0; k<nu; k++) 												
   	   window_pwr = window_pwr + h_sqr[index + k];  
   	   
   	/* select window with greatest signal power */   
   	if(window_pwr > max_power){ 								
	     max_power = window_pwr;
   	   delta = index;
	  } 
	  /* stop if the signal power is now too small */
   	if(window_pwr < (.1* max_power) )						
	      break;
	}
  delta++;  
  
  /* initialize window matrix	*/
	for(i=0; i<nu; i++)														
	   for(j=0; j<Nw; j++)
	   		ipr_sect[i][j] = 0;											
  
 	/* calculate window */	 			
	for(i=delta; i<delta + nu; i++)								
	   for(j=0; ((j<i+1)&&(j<Nw)); j++)
	   		ipr_sect[i-delta][j] = h[i - j];	

	/* Y = (win)'* win	 */
	for(i=0; i<Nw; i++)														
	   for(j=0; j<Nw; j++){
	   		Y[i][j] = 0;	   		
	   		for(k=0; k<nu; k++)
	   			Y[i][j] += ipr_sect[k][i] * ipr_sect[k][j];   			
	   }

	/* initialize wall matrix */	
	for(i=0; i<N; i++)														
	   for(j=0; j<Nw; j++)
	   		ipr_sect[i][j] = 0;											
   
  /* first part of wall */
	for(i=0; i<delta; i++)												
	   for(j=0; ((j<i+1)&&(j<Nw)); j++)
	   		ipr_sect[i][j] = h[i - j];	
  
  /* second part of wall */
	for(i=delta + nu; i<N; i++)										
	   for(j=0; ((j<i+1)&&(j<Nw)); j++)
	   		ipr_sect[i][j] = h[i - j];	
  
  /* convert from float to Q15 data type */
  errorcode = fltoq15 (&h_sqr_float, &gain[0], N);
  /* unpack real elements */
  unpacki(&gain[0], N); 
  /* perform bit-reversal on h_sqr */ 
  cbrev (&gain[0], &gain[0], N/2);
	/* calculate ifft of channel gain */
	rifft (&gain[0], N, 1);        
  /* convert from q15 to float data type */  
  q15tofl(&gain[0], &h_sqr_float, N);
   
    
  /* first non-masked row in Hwall */
  start = delta+nu; 					
		
	/*  create first row of B * wall */  
	for (i=0; i<Nw; i++){
    temp[0][i] = 0;
    for (j=start; j<=i+last; j++)
			temp[0][i] += ifft[j]*ipr_sect[j][i]; 
   }
     
  /* create first column of B*wall */
	for (i=1; i<N; i++){
		temp[i][0] = 0; 
		for (j=start; (j<=i)&&(j<=last); j++)
 			temp[i][0] += ifft[i-j]*ipr_sect[j][0];
		for (j=j; j<=last; j++)       
			temp[i][0] += ifft[j-i]*ipr_sect[j][0];
	}
     
	if(first < delta){	 		 
		diff = delta - first;
		 
		/*  add to first row of B*wall */  
		for (i=1; (i<Nw)&&(i<diff); i++)
    	for (j=first+i; j<delta; j++)
				temp[0][i] += ifft[j]*ipr_sect[j][i];	 
     
  	/* add to first column of B*wall */
		for (i=0; i<N; i++){
			for (j=first; (j<=i)&& (j<delta); j++){
			  subtract[i][j-first] = ifft[i-j]*ipr_sect[j][0];
			  temp[i][0] += subtract[i][j-first];
			} 
			for (j=j; j<delta; j++){
			  subtract[i][j-first] = ifft[j-i]*ipr_sect[j][0];
			  temp[i][0] += subtract[i][j-first];      
			}
		}
		
		/* diagonally include MACs for first:delta elements */
		for(i=0; i<N; i++)
			for(j=1; (j<=diff)&&(j<Nw)&&(i+j<N); j++){
				k = start-i-j;
				if(k < 0)	 
					k = -k; 
				temp[i+j][j] = temp[i+j-1][j-1]-subtract[i][diff-j]
											 +ifft[k]*ipr_sect[start][j]; 
			}
		
		/* include MACs for missed elements in diagonal method */	
		for(i=1; i<diff; i++) 			
			for(k=i+1; k<=diff; k++){
				temp[i][k] = temp[i-1][k-1] - 
										 ifft[delta-i]*ipr_sect[delta-1][k-1] + 
										 ifft[start-i]*ipr_sect[start][k]; 
				
			}				
		
		/* fill in remaining elements */	 
		for (i=1; i<N; i++){
			j = start-i;
			if(j < 0)	 
				j = -j;
			for (k=diff+1; k<Nw; k++)
				temp[i][k] = temp[i-1][k-1] + 
										 ifft[j]*ipr_sect[start][k]; 
		}   
	}     
     
	/* else (first >= delta) */
	else{ 
		/* fill in remaining elements */
		for (i=1; i<N; i++){
			j = start-i;
			if(j < 0)	 
				j = -j;
			for (k=1; k<Nw; k++)
				temp[i][k] = temp[i-1][k-1] + 
										 ifft[j]*ipr_sect[start][k]; 
		}
	}
						                                           
	/* X = (D*Hwall)' * (B*D*Hwall) */	
	/* Calculate for upper half */	                                           
	for (i = 0; i < Nw; i++)
		for (j = 0; j < Nw; j++) {
			X[i][j] = 0;
			for(k=first; (k<delta)&&(k<=last+Nw-1); k++)
				X[i][j] += ipr_sect[k][i] * temp[k][j];  			
		}                               
		
  /* X = (D*Hwall)' * (B*D*Hwall) */
  /* Calculate for lower half */
	for (i = 0; i < Nw; i++)
		for (j = 0; j < Nw; j++) {
			for(k=delta+nu; k<=last+Nw-1; k++)
				X[i][j] += ipr_sect[k][i] * temp[k][j];
			X[i][j] *= 100;			 			
		}
						
	/* Compute inv_of_chol(X) */	
	for (i=0;i<Nw;i++) {
		for (j=i;j<Nw;j++) {
			sum = X[i][j];
			for (k=i-1; k>=0; k--) 
				sum -= X[i][k] * X[j][k];
			if (i == j) {
			  /* X is not positive definite */
				if (sum <= 0)						
						puts("choldc failed\n");                     					
				errorcode = fltoq15 (&sum, &q15, 1);		
				/* Take square root of element */
				oflag = sqrt_16 (&q15, &p[i], 1);	
				q15tofl(&p[i], &p_float[i], 1);			
			}
			else X[j][i] = sum/p_float[i];
		}
	}
	for (i=0;i<Nw;i++) {
		X[i][i] = 1/p_float[i];
		for (j=i+1;j<Nw;j++) {
			sum = 0;
			for (k=i;k<j;k++) 
				sum -= X[j][k]*X[k][i];
			X[j][i] = sum/p_float[j];
		}
	}
	for (i=0;i<Nw;i++)
		for (j=i+1;j<Nw;j++)
			X[i][j] = 0;             
              
	/* calculate inv(X) = X * X' , store in temp*/
	for (i = 0; i < Nw; i++)
		for (j = 0; j < Nw; j++) {
			temp[i][j] = 0;
			for (k = 0; k < Nw; k++)
				temp[i][j] += X[i][k] * X[j][k]; 
			temp[i][j] *= 100;
		} 
		
	/* Calculate Z = inv(X) * Y */        
	for (i = 0; i < Nw; i++)
		for (j = 0; j < Nw; j++) {
			Z[i][j] = 0;
			for (k = 0; k <Nw; k++)
				Z[i][j] += temp[i][k] * Y[k][j];
		}

  /* initialize l */
	for(i=0; i<Nw; i++){
		l[i]= rand_vector[i];       
	}

  /* z = Z*l  */
	for (i = 0; i < Nw; i++){
		z[i] = 0;
		for (k = 0; k < Nw; k++)
			z[i] += Z[i][k] * l[k]; 
	}
  
	for (i = 0; i < Nw; i++)
		z_mag += z[i] * z[i];
 
  /* l = z/(z'*z) */
	for (i = 0; i < Nw; i++)
		l[i] = z[i] / z_mag; 
		
	/* calculate Z*l */
	for (i = 0; i < Nw; i++){
			z[i] = 0;
			for (k = 0; k < Nw; k++)
				 z[i] += Z[i][k] * l[k]; 
	}	              
   /*	calculate l'*(Z*l) */
	for (i = 0; i < Nw; i++)
		lambda += l[i] * z[i];
       
  if(lambda_old > lambda)
  	lambda_diff = lambda_old - lambda;
  else 
  	lambda_diff = lambda - lambda_old;

	/* Power Method Iterations  */	                                                       
  /* while((lambda_diff/lambda) > THD){	  //replaced while w/fixed loop	*/
	for(j=0; j<5; j++){
		count++;
		
		/* if could is odd */		
		if(count%2 ==1){ 
			/*   calculate Z*l   */
			for (i = 0; i < Nw; i++){
				z[i] = 0;
				for (k = 0; k < Nw; k++)
					z[i] += Z[i][k] * l[k]; 
			}
    	z_mag = 0;
			for (i = 0; i < Nw; i++)
				z_mag += z[i] * z[i];
 
  		/*	l = z/(z'*z)		*/
			for (i = 0; i < Nw; i++)
				l[i] = z[i] / z_mag; 		  
			
			lambda_old = lambda;
			lambda = 0;
			
			/*   calculate Z*l   */
			for (i = 0; i < Nw; i++){
				z[i] = 0;
				for (k = 0; k < Nw; k++)
					z[i] += Z[i][k] * l[k];
			}				        
			      
	   	/*		calculate l'*(Z*l)		*/
			for (i = 0; i < Nw; i++)
				lambda += l[i] * z[i];
		
	  	if(lambda_old > lambda)
  			lambda_diff = lambda_old - lambda;
  		else 
  			lambda_diff = lambda - lambda_old;
		}
		/* if count is even */
		else{
    	z_mag = 0;
			for (i = 0; i < Nw; i++)
				z_mag += z[i] * z[i];
 
  		/*	l = z/(z'*z)		*/
			for (i = 0; i < Nw; i++)
				l[i] = z[i] / z_mag; 			
		}  
	}	                                                       
           
  /* DON'T PROFILE CODE BELOW/END OF TEQ CALCULATIONS */
  
	for(i=0; i<Nw; i++)
		w_opt[i] = l[i];
	
	// calculate effective shortened channel response 
	for (i = 0; i < N; i++){
		h_eff[i] = 0;
		subchannel[i] = i;
		for(k = 0; ((k<(i+1))&&(k<Nw)); k++)
			h_eff[i] += h[i-k] * w_opt[k] * 100; 
	}                   
	for (i = 0; i < N; i++)
		h_eff_short[i] = (short)h_eff[i];
  puts("THE END");	
}

